import os
import subprocess as s

# 1) BOXMOX is installed and working
if 'KPP_HOME' in os.environ.keys():
    try:
        res = s.check_call("validate_BOXMOX_installation")
    except:
        raise OSError("Cannot validate BOXMOX installation using validate_BOXMOX_installation!")
else:
    raise OSError('$KPP_HOME not found in environment - is BOXMOX installed?')

# 2) an work directory has been defined and is useable
work_path = "/does/not/exist"
if 'BOXMOX_WORK_PATH' in os.environ.keys():
    work_path = os.environ['BOXMOX_WORK_PATH']
    try:
        if not os.path.isdir(work_path):
            os.makedirs(work_path)
    except:
        raise OSError("Could not create work directory " + work_path)
else:
    raise OSError('$BOXMOX_WORK_PATH not found in environment - set it to a path were BOXMOX can write stuff to.')

if not os.path.isdir(work_path):
    import warnings
    warnings.warn("BOXMOX model unusable - experiment execution disabled.")
else:
    from .experiment import Experiment, ExperimentFromExample, ExperimentFromExistingRun, Namelist, examples, compiledMechs

from .data import InputFile, Output, ConcentrationOutput, RatesOutput, AdjointOutput, JacobianOutput, HessianOutput
from .fluxes import FluxParser

try:
    import matplotlib
    from .plotter import ExperimentPlotter
except:
    import warnings
    warnings.warn('matplotlib not found - plotting disabled.')



class ExperimentPlotter:
    '''
    Plotting routines for BOXMOX experiments.
    '''
    def _plottingTimes(self):
        ptimes = self.__times
        punit  = 'h'
        if (ptimes[-1] - ptimes[0]) < 1.0:
            ptimes *= 3600.0
            punit   = 's'
        return ptimes, punit

    def concentrations(self, specs, scaleFactors = 1e3, colorMap=None, tmin=None, tmax=None):
        '''
        Plot time series of concentrations.
        '''
        import numpy as np
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()

        if isinstance(specs, (str)):
            specs = [ specs ]

        nspecs = len(specs)

        if isinstance(scaleFactors, float):
            scaleFactors = [ scaleFactors for x in range(nspecs) ]

        ptimes, punit = self._plottingTimes()
        if tmin is None:
            tmin=ptimes[0]
        if tmax is None:
            tmax=ptimes[-1]

        if colorMap is None:
            colorMap = plt.cm.brg
        color=iter(colorMap(np.linspace(0,1,nspecs)))

        for scale, spec in zip(scaleFactors, specs):
            c=next(color)
            ax.plot(ptimes, self.__conc[spec] * scale, color=c, label=spec, linewidth=2)
        ax.set_xlabel('Time in '+punit)
        ax.set_xlim( tmin, tmax )

        ax.legend()

    def fluxes(self, spec, fig=None, ax=None,
               plot_legend=True, plot_netflux=True, plot_xaxlabels=True,
               scale_factor=3600.0 * 1000.0, ylim=None,
               xlabel='Time in h', ylabel=r'Flux in ppbv h$^{-1}$', plabel=None, flx_labels = None,
               xaxis=True, yaxis=True):
        '''
        Plot time series of fluxes through a species.
        '''
        import numpy as np
        import matplotlib.pyplot as plt

        if fig is None:
            fig, ax = plt.subplots()

        flxs = self.__flxs.get(spec)

        times   = flxs['time']
        flx     = { x: flxs['values'][x] for x in flxs['values'].keys() if not x == 'time' }

        netflux = np.zeros(len(times))

        for xkey, xvalue in flx.iteritems():
            try:
                ax.plot(times, xvalue * scale_factor, label=flx_labels[xkey]['label'], c=flx_labels[xkey]['color'])
            except:
                ax.plot(times, xvalue * scale_factor, label=xkey)
            netflux += np.array(xvalue)

        if plot_netflux:
            ax.plot(times, netflux * scale_factor, label="Net flux", lw=1, c='k')

        ax.grid( alpha=0.5, color='gray', linestyle='solid' )
        ax.plot(times, np.zeros(len(times)), lw=0.5, alpha=0.5, c='dimgray')
        ax.set( xlim=(times[0],times[-1]), xlabel=xlabel, ylabel=ylabel)
        if not ylim is None:
            ax.set( ylim=ylim )

        if not plabel is None:
            xmin, xmax = ax.get_xlim()
            ymin, ymax = ax.get_ylim()
            ax.text( xmin + 0.05 * (xmax - xmin), ymin + 0.95 * (ymax - ymin), plabel,
                     horizontalalignment='center', verticalalignment='center')

        if plot_legend:
            ax.legend(loc='upper center', fontsize="xx-small", fancybox=True, framealpha=0.75)

        if not plot_xaxlabels:
            ax.tick_params(labelbottom='off')

        return fig

    def __init__(self, output, fluxes):
        self.__conc = output['Concentrations']
        self.__flxs = fluxes
        self.__times = self.__conc['time']



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import pyparsing as pp
import numpy as np

class FluxParser:

    def _init_reac_parser(self):

        #--------------------------------------------------
        # Format of reaction equation:
        # (factor) species + ... --> (factor) species + ...
        #--------------------------------------------------

        plus   = pp.Literal('+')  .suppress()
        arrow  = pp.Literal('-->').suppress()
        spec   = pp.Word( pp.alphanums+'_' ).setParseAction(lambda f: f[0].upper())
        factor = pp.Word( pp.nums+'.' ).setParseAction(lambda f: float(f[0]))

        spec_mem = pp.Optional(factor, default=1.0) + spec + pp.Optional(plus)

        self.reac_parser = pp.Group( pp.OneOrMore(spec_mem) ) + \
                           arrow                              + \
                           pp.Group( pp.OneOrMore(spec_mem) )

    def get (self, spec, min_flux_percentage=0.1, CFACTOR=2.596379e13):

        # Extract sink and source reactions:
        sinks, sources  = {}, {}
        reactants       = []
        for xkey,xvalue in self.rates.iteritems():

            xreac = self.reac_parser.parseString(xkey).asList()

            if spec in xreac[0]:
                sinks[xkey] = xreac
                for xspec in xreac[0]:
                    if type(xspec)==str:
                        reactants.append( xspec )

            elif spec in xreac[1]:
                sources[xkey] = xreac
                for xspec in xreac[0]:
                    if type(xspec)==str:
                        reactants.append( xspec )

            else:
                pass

        # Involved reactants of which concentration is needed:
        reactants = list( set( reactants ) )

        # Example for how to calculate chemical fluxes:
        #      Searched is the flux of species C in the following reaction:
        #      Reaction:  A + B -> 0.3C + 0.7D + 0.6E
        #      Flux    :  [A] * [B] * r * 0.3 = flux_C = 0.3*[C] * 0.7*[D] * 0.6*[E] * r
        #
        #      Where [x] denotes the concentration of x and r is the rate constant of the
        #      present reaction.
        #
        def calc_flux ( spec, reac, conc, rates ):

            flux = {}

            for xkey,xvalue in reac.iteritems():
                xflux = np.ones_like( rates[xkey], dtype=float ) * rates[xkey]

                for ispec in range(1,len(xvalue[0]),2):
                    # The concentrations in the files are ppmv
                    # CFACTOR -> concentrations in mlc/cm3.
                    # The rates are in e.g. cm3/mlc/s -> convert concentrations in
                    # mlc/cm3
                    xconc = conc[xvalue[0][ispec]][:] * CFACTOR
                    # Every species has its own factor (mostly 1.0):
                    xfac  = xvalue[0][ispec-1]

                    xflux *= (xconc**xfac)

                flux[xkey] =  xflux / CFACTOR

                # Divide by the CFACTOR to get the flux in ppmv/s instead of mlc/cm3/s:
                if spec in xvalue[0]:
                    flux[xkey] = -1.0 * xflux / CFACTOR   # sinks are negative
                elif spec in xvalue[1]:
                    # If the reaction is a source reaction the factor of the investigated
                    # species may not be 1.0 ...
                    flux[xkey] =  1.0 * xflux / CFACTOR * xvalue[1][xvalue[1].index(spec)-1]
                else:
                    raise IOError('species <{0:s}> not contained in reaction "{1:s}"'\
                                  .format(spec,xkey))

            return flux

        fluxes = calc_flux(spec,sinks,self.conc,self.rates)
        fluxes.update( calc_flux(spec,sources,self.conc,self.rates) )

        # Remove fluxes that account for less then <min_flux_percentage> of total net flux
        #print fluxes.values()
        netFlux_pos=0.0
        netFlux_neg=0.0

        for indx,val in enumerate(np.array(fluxes.values())[:,0]):
            netFlux_pos += np.sum( [x for x in np.array(fluxes.values())[indx,1:-1] if x>0.0]  )
            netFlux_neg += np.sum( [x for x in np.array(fluxes.values())[indx,1:-1] if x<0.0]  )

        without_negligible_fluxes = {}
        for xkey,xvalue in fluxes.iteritems():
            if np.sum(xvalue[1:-1]) < min_flux_percentage * netFlux_pos and np.sum(xvalue[1:-1]) > min_flux_percentage * netFlux_neg :
                pass
            else:
                without_negligible_fluxes[xkey] = xvalue

        return { 'species': spec, 'time': self.time, 'values': without_negligible_fluxes }

    def __init__(self, output):
        self.valid = False

        self.output = output

        self._init_reac_parser()

        # Output of single BOXMOX run:
        ratesObj    = self.output['Rates']
        self.time   = ratesObj.time
        self.rates  = ratesObj.rates

        self.conc   = self.output['Concentrations']




import os
import subprocess as s
import threading
import shutil
import fnmatch
import datetime
import f90nml

from .data import InputFile, Output, ConcentrationOutput, RatesOutput, AdjointOutput, JacobianOutput, HessianOutput
try:
    from .plotter import ExperimentPlotter
except:
    pass
from .fluxes import FluxParser

workPath=os.environ['BOXMOX_WORK_PATH']
examplesPath=os.path.join(os.environ['KPP_HOME'], "case_studies")

class ExampleData:
    def __getitem__(self, name):
        if name in self.files.keys():
            return InputFile(self.files[name])
        else:
            raise Exception("Input file {:s} does not exist.".format(name))
    def __init__(self, path):
        self.path  = path
        self.files = { x.replace(".csv", ""): os.path.join(self.path, x) for x in os.listdir(self.path) if '.csv' in x }
        self.mech  = 'unknown'
        try:
            with open( os.path.join( self.path, 'mech_used' ) ) as f:
                self.mech = f.readline()
        except:
            pass

compiledMechsPath=os.path.join(os.environ['KPP_HOME'], "compiled_mechs")

compiledMechs = []
try:
    compiledMechs = [ x for x in os.listdir(compiledMechsPath) if (os.path.isdir( os.path.join(compiledMechsPath, x) )) and ("_adjoint" not in x) ]
except:
    print("No compile_mechs directory found")

_examples = { x: ExampleData(os.path.join(examplesPath, x)) for x in os.listdir(examplesPath) if os.path.isdir( os.path.join(examplesPath, x) ) }
examples  = { x: _examples[x] for x in _examples if _examples[x].mech in compiledMechs }

class Namelist:
    '''
    A BOXMOX (Fortran 90) namelist. Getting and setting of values
    works like a dictionary::

       print(nml['tstart'])
       nml['tstart'] = 12.0

    '''
    def __getitem__(self, item):
        if item.lower() in self._namelist['boxmox'].keys():
            return self._namelist['boxmox'][item.lower()]
        else:
            raise Exception('{:s} does not exist in Namelist.'.format(item))
    def __setitem__(self, item, value):
        if item.lower() in self._namelist['boxmox'].keys():
            try:
                self._namelist['boxmox'][item] = value
            except:
                raise Exception('Failed to set {:s} to {:s}'.format(item, str(value)))
        else:
            raise Exception('{:s} does not exist in Namelist'.format(item))
    def __str__(self):
        return ', '.join( [ ('{:s}: {:s}'.format(item, str(self._namelist['boxmox'][item]))) for item in self._namelist['boxmox'].keys() ] )
    def keys(self):
        return self._namelist['boxmox'].keys()
    def write(self, path):
        '''
        Write namelist to file.
        '''
        self._namelist.write(path, 'w')
    def read(self, path):
        '''
        Read namelist from file.
        '''
        self._namelist = f90nml.read(path)
    def __init__(self, path = None):
        self._namelist = None

        self.read(os.path.join(examplesPath, "BOXMOX.nml"))
        if not path is None:
            if os.path.exists(path):
                self.read(path)

class Experiment:
    '''
    A new BOXMOX experiment using a certain mechanism.
    '''
    removeOnDel = True

    def __init__(self, mechanism):
        import uuid

        #: Unique ID
        self.name = str(uuid.uuid4())
        #: Chemical mechanism used
        self.mechanism = mechanism

        #: Absolute file system path of the experiment
        self.path = os.path.join(workPath, self.name)

        #: Input files
        self.input       = {}
        #: Output files
        self.output      = {}

        self.plotter     = None
        self.fluxes      = None

        #: Log file of the last run
        self.log         = None

        self.pid         = None
        self.thread      = None

        self.__new()

        #: Namelist
        self.namelist = Namelist(os.path.join(self.path, 'BOXMOX.nml'))

        self.version  = self._determineVersion(self.path)

    def _determineVersion(self, path):
        version     = 1.8
        versionFile = os.path.join(path, 'VERSION')
        if os.path.exists(versionFile):
            try:
                with open(versionFile, 'r') as f:
                    line = f.readline().rstrip()
                    line = ".".join(line.split(".")[0:min(2, len(line.split(".")))])
                    # for development only:
                    if line == "__BOXMOX_VERSION__":
                        line = 1.8
                    version = float(line)
            except:
                version = 1.8
                pass
        return version

    def _populateInput(self):
        inputFileNames = [f for f in os.listdir(self.path) if f.endswith('.csv')]
        self.input = { x.replace('.csv', '') : InputFile(os.path.join(self.path, x)) for x in inputFileNames }

    def __new(self):
        try:
            s.call(["new_BOXMOX_experiment", self.mechanism, self.path])

            self._populateInput()
        except s.CalledProcessError as e:
            raise Exception("Creating new experiment failed: " + str(e.output))

    def addInputFromFile(self, name, fname, overwrite=False):
        '''
        Add input from file
        '''
        if name in self.input.keys() and not overwrite:
            raise Exception("Input file {:s} already exists.".format(name))
        self.input[name] = InputFile(fname)

    def addInput(self, name, inp, overwrite=False):
        '''
        Add input from InputFile object
        '''
        if name in self.input.keys() and not overwrite:
            raise Exception("Input file {:s} already exists.".format(name))
        self.input[name] = inp

    def _get_mechanism(self):
        mech = None

        # see if we have an executable
        exes = [ x for x in os.listdir(self.path) if '.exe' in x ]
        if len(exes) == 1:
            mech = exes[0].replace('.exe', '')

        # if not, use an output file
        if mech is None:
            outs = [ x for x in os.listdir(self.path) if x.endswith('.conc') ]
            if len(outs) == 1:
                mech = outs[0].replace('.conc', '')

        if mech is None:
            import warnings
            warnings.warn('Could not determine mechanism in BOXMOX experiment path: ' + self.path)

        return mech

    @property
    def running(self):
        '''
        Check if simulation is running (if started with asynchronous=True).
        '''
        if self.thread is None:
            return False
        else:
            return self.thread.isAlive()

    def _run(self, dumbOutput=False):
        try:
            self.pid = s.Popen([ "./" + self.mechanism + ".exe" ], stdout=s.PIPE, cwd=self.path)
            self.pid.wait()
            self._pp_run(dumbOutput=dumbOutput)
        except Exception as e:
            failDir = self.path + "_failed_at_{:s}".format(datetime.datetime.strftime(datetime.datetime.now(), "%Y%m%d%H%M%S"))
            self.archive( failDir )
            raise Exception('BOXMOX integration failed: {:s}'.format(str(e)))
        return

    def run(self, dumbOutput=False, asynchronous=False):
        '''
        Run BOXMOX experiment

        :param bool dumbOutput: Only load references to output files, instead of loading data? (faster, but not usable for analysis)
        :param bool asynchronous: Run asynchronous?
        '''
        if asynchronous:
            if self.running:
                return
        
        self.namelist.write(os.path.join(self.path, 'BOXMOX.nml'))

        for type in self.input:
            with open( os.path.join(self.path, type + '.csv'), 'w') as f:
                self.input[type].write(f)

        pwd = os.getcwd()
        os.chdir(self.path)

        if asynchronous:
            self.thread = threading.Thread(target=self._run, args=(dumbOutput,))
            self.thread.start()
            os.chdir(pwd)
            return
        else:
            self._run(dumbOutput=dumbOutput)
            os.chdir(pwd)

    def _pp_run(self, dumbOutput=False):
        try:
            self.log = []
            for line in iter(self.pid.stdout.readline, b''):
                self.log.append(line.replace('\n',''))
        except:
            pass

        if dumbOutput:
            self._populateDumbOutput()
        else:
            self._populateOutput()

    outputTypes = {  'Concentrations': { 'ending': '.conc',     'class': ConcentrationOutput },
                     'Rates':          { 'ending': '.rates',    'class': RatesOutput },
                     'Jacobian':       { 'ending': '.jacobian', 'class': JacobianOutput },
                     'Hessian':        { 'ending': '.hessian',  'class': HessianOutput },
                     'Adjoint':        { 'ending': '.adjoint',  'class': AdjointOutput } }

    def _populateDumbOutput(self, filename="*"):
        for tname in self.outputTypes:
            otype = self.outputTypes[tname]

            fnames = [ x for x in os.listdir(self.path) if x.endswith(otype['ending']) and fnmatch.fnmatch(x, filename+otype['ending']) ]
            if len(fnames) > 0:
                if len(fnames) > 1:
                    raise Exception('Multiple output files found in path {:s} for type {:s}: {:s}'.format(self.path, tname, ', '.join(fnames)))
                self.output[tname] = Output(os.path.join(self.path, fnames[0]))

    def _populateOutput(self, filename="*"):
        for tname in self.outputTypes:
            otype = self.outputTypes[tname]

            fnames = [ x for x in os.listdir(self.path) if x.endswith(otype['ending']) and fnmatch.fnmatch(x, filename+otype['ending']) ]
            if len(fnames) > 0:
                if len(fnames) > 1:
                    raise Exception('Multiple output files found in path {:s} for type {:s}: {:s}'.format(self.path, tname, ', '.join(fnames)))
                self.output[tname] = otype['class'](os.path.join(self.path, fnames[0]))

        if 'Concentrations' in self.output.keys() and 'Rates' in self.output.keys():
            self.fluxes  = FluxParser(self.output)
        if 'Concentrations' in self.output.keys():
            try:
                self.plotter = ExperimentPlotter(self.output, self.fluxes)
            except:
                pass

    def plot(self, specs, **kwargs):
        '''
        Shortcut to plotter.concentrations for quick timeline plotting
        '''
        if not self.plotter is None:
            self.plotter.concentrations(specs, **kwargs)
        else:
            import warnings
            warnings.warn('Plotter not useable - is matplotlib installed?')

    def archive(self, archive_path):
        '''
        Save BOXMOX experiment to archive path
        '''
        try:
            shutil.copytree(self.path, archive_path)
        except Exception as e:
            raise Exception('Archiving experiment at {:s} to {:s} did not work: {:s}'.format(self.path, archive_path, str(e)))

    def __del__(self):
        def deldidntwork(fun, path, excinfo):
            raise IOError('Could not remove experiment path {:s}: {:s}'.format(path, excinfo))
        if self.removeOnDel:
            shutil.rmtree(self.path, onerror=deldidntwork)

class ExperimentFromExample(Experiment):
    '''
    A BOXMOX experiment based on an example from the BOXMOX installation.
    '''
    def __init__(self, example):
        import uuid

        self.name = str(uuid.uuid4())
        self.example = example

        self.path = os.path.join(workPath, self.name)

        self.log         = None

        self.pid         = None
        self.thread      = None

        self.input       = {}
        self.output      = {}

        self.plotter     = None
        self.fluxes      = None

        self.__new()

        self.namelist = Namelist(os.path.join(self.path, 'BOXMOX.nml'))
        self.mechanism = self._get_mechanism()

        self.version  = self._determineVersion(self.path)

    def __new(self):
        try:
            s.check_output(["new_BOXMOX_experiment_from_example", "-f", self.example, self.path], stderr=s.STDOUT)
            self._populateInput()
        except s.CalledProcessError as e:
            raise Exception("Creating new experiment failed: " + str(e.output))

class ExperimentFromExistingRun(Experiment):
    '''
    A BOXMOX experiment using an existing experiment.
    '''
    removeOnDel = False

    def __init__(self, path, filename="*"):
        '''
        :param str path: path to the existing run
        :param str filename: file name to look for (e.g., for MOZART_4.conc, MOZART_4.rates ...,
                            filename would be "MOZART_4")
        '''
        import uuid

        self.name = str(uuid.uuid4())

        self.path = os.path.abspath(path)

        self.log         = None

        self.pid         = None
        self.thread      = None

        self.input       = {}
        self.output      = {}

        self.plotter     = None
        self.fluxes      = None

        self.namelist = Namelist(os.path.join(self.path, 'BOXMOX.nml'))
        self.mechanism = self._get_mechanism()

        self.version  = self._determineVersion(self.path)

        self._populateInput()
        self._populateOutput(filename=filename)







import os
import sys
import shutil
from io import StringIO
import csv

import numpy as np

class InputFile:
    '''
    A generic BOXMOX input file (>= v 1.7). Getting and setting of values
    works like a dictionary::

       print(inp['O3'])
       inp['O3'] = 0.040

    '''
    @property
    def nvar(self):
        '''
        Number of variables.
        '''
        return len( [ x for x in self.keys() ] )
    @property
    def nanc(self):
        '''
        Number of ancillary variables.
        '''
        return len( [ x for x in self.anc.keys() ] )
    def __getitem__(self, item):
        return [ self._data[i] for i in item ] if isinstance(item, list) else self._data[item]
    def __setitem__(self, item, values):
        if isinstance(item, list):
            for i, v in zip(item, values):
                self._data[i] = v
        else:
            self._data[item] = values
    def keys(self):
        return self._data.keys()

    def read(self, fpath):
        '''
        Read input file from path.
        '''
        with open(fpath, 'r') as f:
            nvar         = int(f.readline().replace(',', ''))
            nanc         = int(f.readline().replace(',', ''))
            self.timeFormat   = int(f.readline().replace(',', ''))

            curpos = f.tell()
            try:
                dialect = csv.Sniffer().sniff(f.read(1048576), delimiters=";, ")
                dialect.skipinitialspace = True
                f.seek(curpos)
                spamreader = csv.reader(f, dialect)
            except:
                # could not determine dialect, falling back to default.
                f.seek(curpos)
                spamreader = csv.reader(f, skipinitialspace = True, delimiter=" ")
            # twice as fast as np.genfromtxt(..., names=True)
            hdr = next(spamreader)
            dmp = []
            for row in spamreader:
                dmp.append( tuple(map(float, row)) )

        ntime = 0
        if not self.timeFormat == 0:
            self.timeVar  = hdr[0]
            self.time     = [ x[0] for x in dmp ]
            ntime = 1
        self.anc       = { hdr[i]: [ x[i] for x in dmp ] for i in range(ntime, ntime+nanc) }
        self._data     = { hdr[i]: [ x[i] for x in dmp ] for i in range(ntime+nanc, len(dmp[0])) }

    def __str__(self):
        f = StringIO()
        self.write(f)
        return(f.getvalue())

    def write(self, f=sys.stdout):
        '''
        Write to <f>. <f> can be file handle or other connection. Defaults to sys.stdout.
        '''

        f.write('{0:1d}'.format(self.nvar)                  +'\n')
        f.write('{0:1d}'.format(self.nanc)                  +'\n')
        f.write('{0:1d}'.format(self.timeFormat)            +'\n')

        data_names = [ key for key in self.keys()     ]
        anc_names  = [ key for key in self.anc.keys() ]
        hdr_line   = '{0:s}\n' .format(' '.join(anc_names + data_names))
        if str(self.timeFormat) != "0" :
            hdr_line = self.timeVar + ' ' + hdr_line
        f.write(hdr_line)

        if isinstance(self.time, list):
            for itime, xtime in enumerate(self.time):
                line = [ xtime ] + [ self.anc[key][itime] for key in anc_names ] + [ self._data[key][itime] for key in data_names ]
                f.write(' '.join('{0:e}'.format(x) for x in line) + '\n')
        else:
            def _data_fix(key):
                if isinstance(self._data[key], list):
                    return self._data[key][0]
                else:
                    return self._data[key]
            data_line = ' '.join( '{0:e}'.format( float(self.anc[x])) for x in self.anc.keys() ) + ' ' + ' '.join( '{0:e}'.format( float(_data_fix(x))) for x in self.keys() ) + '\n'
            if not self.time is None:
                data_line = '{0:e}'.format(float(self.time)) + ' ' + data_line
            f.write(data_line)

    def __init__(self, fpath=None, version=1.8):
        #: Time format (0: constant, 1: seconds since start, 2: hour of diurnal cycle)
        self.timeFormat = 0
        self.timeVar    = 'time'
        self.time       = None

        self.anc        = {}

        self._data     = {}

        self.version    = version

        #: File path of the underlying file (if it exists (yet))
        self.fpath       = fpath
        if not self.fpath is None:
            self.read(self.fpath)

class Output(object):
    '''
    Base output class. Simply an object that knows the output file name,
    can copy and remove it.
    '''
    def rm(self):
        '''
        Remove output file
        '''
        try:
            os.remove(self.fpath)
        except Exception as e:
            raise("Removing output file did not work: {:s}".format(str(e)))
    def write(self, f=sys.stdout):
        '''
        Write contents of Output into file

        :param File f: file object or stream where to write to, defaults to sysout.
        '''
        with open(self.fpath, 'r') as i:
            for line in i:
                f.write(line)

    def copy(self, target_path):
        '''
        Copy output file to target_path

        :param str target_path: location where to copy the output file to.
        '''
        #TODO don't depend on the input file...
        shutil.copy(self.fpath, target_path)
    def __init__(self, fpath):
        '''
        :param str fpath: Output file path
        '''
        self.fpath = fpath

class ConcentrationOutput(Output):
    '''
    Concentration time series
    '''
    def __getitem__(self, items=None):
        if items is None:
            items = self.vars
        return self.data[items]
    def simplified(self, items=None):
        '''
        Get simple 2D NumPy array of concentration vs. time for <items>
        '''
        if items is None:
            items = self.vars
        if isinstance(self.vars, list):
            out = np.zeros( ( len(self.times), len(items) ) )
            for i in range(len(items)):
                out[:,i] = self.data[items[i]]
        else:
            out = self.data
        return out
    def __str__(self):
        fout = StringIO()
        with open(self.fpath, 'r') as f:
            fout.write(f.read())
        return fout.getvalue()

    def __init__(self, fpath, vars=None):
        self.fpath = fpath
        with open(fpath, 'r') as f:
            # not using mygenfromtxt as we know the file format and this is way faster...
            spamreader = csv.reader(f, skipinitialspace = True, delimiter=" ")
            hdr = next(spamreader)
            dat = []
            for row in spamreader:
                dat.append( tuple(map(float, row)) )
        self.data = np.array(dat, dtype=[(_, float) for _ in hdr])
        self.times = self.data['time']
        if not vars is None:
            self.data = self.data[vars]
            self.vars = vars
        else:
            self.vars = list(self.data.dtype.names)

class RatesOutput(Output):
    '''
    Reaction rate time series
    '''
    def __init__(self, fpath):
        self.fpath = fpath
        with open(self.fpath, 'r') as f:
            # First line contains the number of reactions:
            num_of_reacs = int( f.readline() )

            # Skip reaction lines:
            reacStr = []
            for ireac in range( num_of_reacs ):
                reacStr.append( f.readline().strip() )

            # Load rate constants ------------------------------------------------------
            # Every 10000 columns a line break is made ...
            # e.g. column names lines with 30000 reactions (data structure is the same):
            #              line 1 ->   time     1     2     3  ...  10000
            #              line 2 ->  10001 10002 10003 10004  ...  20000
            #              line 3 ->  20001 20002 20003 20004  ...  30000
            #              and so on ...
            rateFormat = int(num_of_reacs/10000 + 1)

            for x in range( rateFormat ):
                # Column names are not needed ...
                _ = f.readline()

            # Length :: number_of_timesteps * rateFormat:
            ratesRaw = [list(map(float, xline.strip().split())) for xline in f.readlines()]

        # "Unfold" the raw data:
        ratesUnfold = ratesRaw[ ::rateFormat]
        for x in range( 1, rateFormat, 1 ):
            ratesHelp   = ratesRaw[x::rateFormat]
            ratesUnfold = [ratesUnfold[ir]+xr for ir,xr in enumerate(ratesHelp)]

        rates = np.array( ratesUnfold )

        # The 0'th row just contains zeros; here the rate constant of the 1'st row is
        # assumed:
        rates[0,:] = rates[1,:]

        # Extract time
        time    = rates[:,0]
        # The 0'th time step has been deleted (line with "rates[0,:] = rates[1,:]")...
        time[0] = 0.0

        # Create output dictionary:
        ratesDict = {xreac:rates[:,ireac+1] for ireac,xreac in enumerate(reacStr)}

        self.rates = ratesDict
        self.time  = time

class JacobianOutput(Output):
    '''
    Jacobian time series
    '''
    def __init__(self, fpath):
        self.fpath = fpath

class HessianOutput(Output):
    '''
    Hessian time series
    '''
    def __init__(self, fpath):
        self.fpath = fpath

class AdjointOutput(Output):
    '''
    Adjoint
    '''
    def __getitem__(self, item1, item2):
        return self.matrix[item1, item2]
    def sensitivity ( self, state, obs ):
        '''
        A BOXMOX adjoint output file is taken and the matrix of the adjoints of
        the species in the state vector is loaded (<adj_mat>: with the function
        read_adjoint).
        With <adj_mat> (= dX(t_final)/dY(t_initial)) the adjoint sensitivities
        of the species in the state vector to the observed species are
        calculated by using the chain rule.

        Calculation of adjoint sensitivities by using adjoints:

        dX(t_final)/dY(t_final) = dX(t_final)/dY(t_initial) * (dY(t_final)/dY(t_initial))^-1

        where:   X  state species
                 Y  observed species

        :param str fname: BOXMOX adjoint output file
        :param str/list state: State vector; single species or list of species which exists/exist in the BOXMOX run
        :param str/list obs: Observed species; single species or list of species which exists/exist in the BOXMOX run

        :return: adjoint sensitivity of the species in the state vector to the observed species
        :rtype: numpy.ndarray (dim_obs, dim_state)
        '''

        # If <state>/<obs> is a single species and not a list a <state>/<obs>-list will be
        # generated
        state = state if type(state)==list else [state]
        obs   = obs   if type(obs  )==list else [obs  ]

        # Get dimensions:
        dim_obs, dim_state = len(obs), len(state)

        # Calculate adjoint sensitivity ::  dX(t_final)/dY(t_final) =
        #                          = dX(t_final)/dY(t_final) * (dY(t_final)/dY(t_initial))^-1
        # shape :: (dim_obs,dim_state)
        adj_sen = np.zeros( (dim_obs,dim_state), dtype=float )
        for iobs,xobs in enumerate( obs ):
            ind_obs_state = state.index( xobs )

            for istate in range( dim_state ):
                # Prevent division by zero ...
                if self.matrix[ind_obs_state,ind_obs_state]!=0.0:
                    # dX(t_final)/dY(t_final) =
                    #            = dX(t_final)/dY(t_initial) * (dY(t_final)/dY(t_initial))^-1
                    adj_sen[iobs,istate] = self.matrix[istate       ,ind_obs_state] *     \
                                           self.matrix[ind_obs_state,ind_obs_state]**-1
                else:
                    adj_sen[iobs,istate] = 0.0

        return adj_sen

    def __init__(self, fpath, state=None):
        self.fpath = fpath
        # Read adjoint file ...
        # Shape of file:
        # Line 0: species names
        # Column 0: "d/d"
        # Column 1: species names
        # (Line :: 1-end, column :: 2-end): data
        with open(self.fpath, 'r') as f:
            # first line species names; entry 0 :: "spc" -> "[1:]"
            spec_names = f.readline ().split()[1:]
            com_data   = f.readlines()

        # Split lines:
        data = []
        for line in com_data:
            # column  0 :: "d/d"; column 1 :: species names -> "[2:]"
            data.append( line.split()[2:] )

        if state is None:
            state = spec_names
        # If <state> is a single species and not a list a <state>-list will be generated
        state = state if type(state) == list else [state]

        # Get dimension:
        dim_state = len(state)

        # Get the indices of the columns of the species in <state>
        # (These indices are equal to the indices of the lines of the species in <state>):
        state_ind = []
        for xstate in state:
            state_ind.append( spec_names.index( xstate ) )

        # Read the adjoints of the species in the state vector
        # shape :: (dim_state,dim_state)
        adj_mat = np.zeros((dim_state,dim_state), dtype=float)
        for istate_ind,xstate_ind in enumerate(state_ind):
            for jstate_ind,ystate_ind in enumerate(state_ind):
                try:
                    adj_mat[istate_ind,jstate_ind] = float( data[xstate_ind][ystate_ind] )
                except Exception as e:
                    print("Problem reading adjoint value <{:s}>: {:s}".format(data[xstate_ind][ystate_ind], str(e)))

        self.matrix = adj_mat



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['boxmox']

package_data = \
{'': ['*']}

install_requires = \
['f90nml>=1.3', 'numpy>=1.2', 'pyparsing>=3']

setup_kwargs = {
    'name': 'boxmox',
    'version': '1.2.5',
    'description': 'BOXMOX python interface',
    'long_description': "# BOXMOX\n\n``boxmox`` is the Python wrapper for the chemical box model BOXMOX (a standalone\nC/Fortran executable).\n\n## Installation\n\n### BOXMOX model needs to be installed\n\nThe BOXMOX chemical box model needs to be installed and the ``KPP_HOME`` environment variable has to be set. Download and instructions are our website at https://mbees.med.uni-augsburg.de/boxmodeling.\n\n### Environment variable needs to be set\n\nAdditionally, ``boxmox`` needs a path to write temporary model results\nto, given through the environment variable ``BOXMOX_WORK_PATH``. This directory needs to be accessible and writeable by the user. Set it in your environment, e.g., through:\n\n```\nexport BOXMOX_WORK_PATH=/where/you/want/boxmox/to/write/stuff/to/\n```\n\nRemember to close the shell and log in again for these changes to take effect.\n\n## Contributing\n\nWe are looking forward to receiving your [new issue report](https://mbees.med.uni-augsburg.de/gitlab/mbees/boxmox_pypackage/-/issues/new).\n\nIf you'd like to contribute source code directly, please [create a fork](https://mbees.med.uni-augsburg.de/gitlab/mbees/boxmox_pypackage),\nmake your changes and then [submit a merge request](https://mbees.med.uni-augsburg.de/gitlab/mbees/boxmox_pypackage/-/merge_requests/new) to the original project.\n\n# Changelog\n\n## 1.2.5 (2022-03-14)\n\n- Release on PyPI\n\n## 1.2.0 (2022-03-08) (not released)\n\n- Updates to be compatible with BOXMOX 1.8\n\n## 1.1.0 (2020-09-16)\n\n- Python 3 compatible \n\n## 1.0.0 (2017-12-19)\n\n- Peer-reviewed version to be published in Knote et al., GMD\n\n## 0.1.0 (2017-08-12)\n\n- Initial release\n",
    'author': 'Christoph Knote',
    'author_email': 'christoph.knote@med.uni-augsburg.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://mbees.med.uni-augsburg.de/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3,<4',
}


setup(**setup_kwargs)
